# Function to read in Asteroid data
# Returns an object with two entries: $session (information applying to the whole test, e.g. subject ID)
# and $trials (data-frame with each trial on a different row)
readAsteroidFile = function(filename) {
  # Outputs a list with two values:
  # session_data -- data about the whole test: subject ID, version etc
  # trial_data_subject -- data about the individual trials
  
  library(data.table)
  library(stringr)
  library(lubridate)
  
  # prepare empty data frame
  session_data = data.frame(
    code = NA,
    pptID = NA,
    datetime = NA,
    date = NA,
    time = NA,
    version = NA,
    OS = NA,
    deviceID = NA,
    cutscenes = NA,
    tracking = NA,
    theme = NA,
    model = NA,
    nalter = NA,
    start = NA,
    threshold = NA,
    ntrials = NA,
    finalarcsec = NA
  )
  
  # Read in the header info
  test = read.table(
    filename,
    header = FALSE,
    nrows = 4,
    sep = ',',
    fill = TRUE,
    as.is = c(1, 2, 3, 4, 5, 6)
  )
  # as.is says: read all files in in teh format they are, ie do not convert to factors
  
  # Critically, see what version of Asteroid this was:
  session_data$versionchar =  as.character(test[1,3])  
  # Now convert it into a number so I can do "version before/after"
  tmp = str_extract_all(test[1,3],"\\(?[0-9,.]+\\)?")[[1]]
  # Split it up by dots:
  g=unlist(strsplit(tmp,'[.]'))
  # Now paste it back together keeping only the first dot
  tmp = paste(g[1],'.',paste(g[2:length(g)],collapse=''),sep='')
  session_data$version =  as.numeric(tmp)  
  session_data$pptsexstr = "Unknown"
  # Account for differences in the number of header rows between versions:
  if (session_data$version<1.0)
  {
    nheaderrows = 5 ; # number of header rows
    rowtheme = 3; # row number with cutscenes, theme etc
    rownAFC = 4; # row number with n in nAFC etc
  }   else
  {
    nheaderrows = 6 ; # number of header rows
    rowtheme = 4; # row number with cutscenes, theme etc
    rownAFC = 5; # row number with n in nAFC etc
    # Also this includes pt info
    session_data$pptDOB = dmy(test[3, 2])
    session_data$pptsexstr = test[3, 4]
  }
  # Early versions output data-files with 10 columns, later ones with 22:
  if (session_data$version >= 0.939)  { ncols = 22 }   else { ncols = 10 }
  
  # Put this heading info into the data frame:
  # First, if the file records a pptID internally, use that:
  session_data$pptID = as.character(test[2, 2])
  # But if it's empty, take it from the header:
  if (nchar( session_data$pptID )==0)
  {
    tmp = strsplit(filename, "-") # split at -
    # the second one is pptID, e.g. "trialData-005-dec1c806bc85de4c58776276cae87687-17-12-2019-Time-19H-29M-37S.csv" , then tmp[[1]][2] = "005"
    # But strip leading zeros for consistency with the small-dots version (arggh!)
    session_data$pptID = str_remove(tmp[[1]][2], "^0+") 
  }
  session_data$OS = as.character(test[1, 5])
  session_data$datetime = mdy_hms(test[1, 1])
  
  
  # also record date and time separately
  tmpD = strsplit(test[1, 1], " ")
  session_data$date  = tmpD[[1]][1]
  
  session_data$time  = tmpD[[1]][2]
  
  session_data$deviceID = test[1, 7]
  cutscene = test[rowtheme, 2]
  if (is.logical(as.logical(cutscene)))
    session_data$cutscenes = as.logical(cutscene)
  else 
    session_data$cutscenes = as.logical(as.numeric(cutscene))
  tracking = test[rowtheme, 4]
  if (!is.na(as.logical(tracking)))
    session_data$tracking = as.logical(tracking)
  else 
    session_data$tracking = as.logical(as.numeric(tracking))
  
  session_data$theme = test[rowtheme, 6]
  session_data$model = as.character(test[rowtheme, 8])
  session_data$nalter = test[rownAFC, 2]
  session_data$StartingThreshold = test[rownAFC, 4]
  
  

  

  # Read in data from the file, keeping the header names written in:
  filedata = fread(
    filename,
    sep = ',',
    nrows = -1,
    header = TRUE,
    skip = nheaderrows,
    select = c(1:ncols),
    fill = TRUE,
    check.names = TRUE
  )
  # Rename some columns:
  # Remove all .s in column names
  colnames(filedata) = str_replace_all(colnames(filedata), "\\.", "")
  # do some more tweaking to make each word begin with a capital:
  colnames(filedata)[1] = "TrialNumber"
  colnames(filedata)[2] = "CorrectAnswer"
  colnames(filedata)[3] = "StimulusClicked"
  colnames(filedata)[5] = "ParallaxArcsec"
  colnames(filedata)[7] = "SecondsTaken"
  
  # Take the final threshold to be the last value either tested or proposed as the next trial:
  nrows = nrow(filedata)
  if (nrows == 1)
  {
    return(NULL)
  }
  
  session_data$finalarcsec = filedata$ParallaxArcsec[nrows]
  
  # Special situation for this dataset -check to see if there is a corrected value
  # This would have been created by CorrectData.m, calling CorrectViewingDistanceEror.m, to account for the fact
  # that the viewing distance was actually at 50cm.
  correctedval = filedata[nrows,7]
  if (!is.na(correctedval))
  {
        session_data$finalarcsec_beforecorrection = session_data$finalarcsec;
        session_data$finalarcsec = correctedval;
  }
        
  
  # Remove final row that just reports the final result:
  if (filedata[nrows, 4] >= "Final")
  {
    filedata = filedata[1:(nrows - 1), ]
    nrows = nrows - 1
  }
  
  
  # rename this data-frame from filedata to "trial_data_subect":
  trial_data_subject = as.data.frame(filedata)
  
  
  # adjust trial number (starts at 0 in excel files)
  trial_data_subject$TrialNumber = trial_data_subject$TrialNumber + 1
  # remove duplicate trials
  trial_data_subject = trial_data_subject[!duplicated(trial_data_subject), ]
  
  # Write in null eyetracking data if needed
  Ufields = c(
    "U_STATUS_NO_ACTION",
    "U_EXCESS_ACCELERATION",
    "U_CAM_CAPTURE_FAIL",
    "U_CANNOT_LOCATE_SYMBOL",
    "U_LOCATED_SYMBOL",
    "U_CANNOT_LOCATE_EYES",
    "U_LOCATED_EYES"	,
    "U_LOCATED_BOTH",
    "U_COMPUTED_DIST_SYMBOL_LOW_QUAL",
    "U_COMPUTED_DIST_SYMBOL",
    "U_COMPUTED_DIST_EYES_LOW_QUAL",
    "U_COMPUTED_DIST_EYES",
    "U_COMPUTED_DIST_SYMBOL_AND_EYES"
  )
  
  for (U in Ufields)
  {
    N = str_replace(U, "U_", "N_")
    totalfield = paste("total", str_replace(U,"U_",""), sep = "_")
    # U is for example "U_LOCATED_EYES": cumulative total of the number of times that status was logged on this and all previous trials.
    # N is then "N_LOCATED_EYES", the total number of times that status was logged on this trial.
    # totalfield is "total_LOCATED_EYES", which we will write into session as the total number of times that status was logged on all trials    
    
    if (any(colnames(trial_data_subject) == U) == FALSE)
    {
      # Data-file did not contain this field; write in a null
      trial_data_subject[, U] = NA
      trial_data_subject[, N] = NA
      # Get the total from the Trackingstatus field
      session_data[totalfield] = sum(trial_data_subject$Trackingstatus == U)
    }
    else
    {
      # If there was a field like this, write in its total to the session structure:
      session_data[totalfield] = trial_data_subject[nrows, U]
      # NB in the files, the U-fields report cumulative number of frames, so to get the total, just look at the last row
      # The "U_" columns report cumulative totals
      # Make a new column which reports the number of frames on this trial
      trial_data_subject[1, N] = trial_data_subject[1, U]
      if (nrows>=2) # NB unlike in MAtlab, [2:1] will return [2,1] not an empty array
        {
          trial_data_subject[2:nrows, N] = trial_data_subject[2:nrows, U] - trial_data_subject[1:(nrows -1), U]
      }
    }
  }
  # FInd out the total number of frames for which track-status was recorded on this trial:
  Nfields = colnames(trial_data_subject)[str_detect(colnames(trial_data_subject), 'N_')]
  trial_data_subject$Ntrackstatus = rowSums(trial_data_subject[, Nfields])
  # Remove  the Trackingstatus field which is only present in the earlier versions:
  trial_data_subject$Trackingstatus = NULL
  
  # finished reading in file and getting trial_data_subject
  
  # Write in some more summary data to the session structure:
  
  # Duration of test run (without animations)
  session_data$duration_alltr = sum(trial_data_subject$SecondsTaken)
  session_data$ntrials = nrow(trial_data_subject)
  
  # Information about viewdist algorithm. Convert totals to proportions
  # Get the column names of session_data which begin with "total":
  totalfields = grep("total+",colnames(session_data),value=TRUE)
  totalframes = sum(as.numeric(session_data[,totalfields]), na.rm = FALSE)
  # If any values are missing, totalframes will be NA. 
  # OK now add on corresponding fields storing the fraction:
  for (fld in totalfields)
  {
    session_data[str_replace(fld,"total","frac")] = session_data[fld] / totalframes
  }
  # Record what file was used
  session_data$filename = basename(filename)
  
  trial_data_subject = as.data.frame(trial_data_subject)
  
  
  # Finally, what to return from this function:
  return(list(session = session_data, trials = trial_data_subject))
}
